module Exam.Exercise1.SemanticsBigStep where

open import Exam.Exercise1.Syntax
open import Exam.Exercise1.Substitution

-- First we define which expressions we consider to be values.

-- For the small-step semantics we defined values as a property of terms:
--
--   data Value : Term → Set where
--     V-ƛ    : Value (ƛ x ⇒ M)
--     V-zero : Value `zero
--     V-suc  : Value M → Value (`suc M)
--
-- For the big-step semantics it is more convenient to define values
-- independently of terms and then define an injection ⌞_⌟ which converts values
-- to the terms they correspond to:

data Value : Set where
  ƛ_⇒_  : Id → Term → Value
  `zero : Value
  `suc_ : Value → Value

⌞_⌟ : Value → Term          -- You can input ⌞ with \cl and ⌟ with \clr.
⌞ ƛ x ⇒ M ⌟ = ƛ x ⇒ M       -- Note that we used the same constructor names for
⌞ `zero ⌟   = `zero         -- Terms and Values, so the `zero on the left side
⌞ `suc V ⌟  = `suc ⌞ V ⌟    -- has a different type than the `zero on the right.

-- The big-step semantics is defined as a relation between Terms and Values.
--
-- With the definition of Value used in this file, we can express this relation
-- directly as
--
--   _⇓_ : Term → Value → Set
--
-- whereas with the definition from the small-step semantics, we would need to
-- use an existential type to glue a term M together with a proof that M is
-- a value:
--
--   _⇓_ : Term → ∃[ M ] (Value M) → Set

-- Without further ado, here is the definition of the big-step semantics:

infix 4 _⇓_

data _⇓_ : Term → Value → Set where

  ⇓-ƛ : ∀ {x M} →            -- A ƛ-function is already a value, so it evaluates
                             -- to itself.
    ---------------------
    (ƛ x ⇒ M) ⇓ (ƛ x ⇒ M)

  ⇓-zero :                   -- The same goes for the `zero term.

    -------------
    `zero ⇓ `zero

  ⇓-suc : ∀ {M V} →          -- To evaluate the successor of a term M, we have
                             -- to evaluate M to some value V, and then return
    M ⇓ V →                  -- the successor of V.
    -------------------
    (`suc M) ⇓ (`suc V)

  ⇓-· : ∀ {x M M' N V V'} →  -- To evaluate a function application (M · N), we
                             -- first have to evaluate M to a function value (ƛ x ⇒ M'),
    M ⇓ (ƛ x ⇒ M') →         -- and N to some value V'.
    N ⇓ V' →                 -- Then we use substitution to plug the argument
    M' [ x := ⌞ V' ⌟ ] ⇓ V → -- value V' into the function body M', and evaluate
    ------------------------ -- this modified body to our final value V.
    (M · N) ⇓ V

  -- There are two ways to evaluate a case expression on a term M:

  ⇓-case-zero : ∀ {M M₁ M₂ V₁} →           -- if M evaluates to `zero, then
                                           -- we evaluate the first branch M₁ and
    M ⇓ `zero →                            -- return its value V₁
    M₁ ⇓ V₁ →
    -------------------------------------
    (case M [zero⇒ M₁ |suc x ⇒ M₂ ]) ⇓ V₁

  ⇓-case-suc : ∀ {M M₁ M₂ V V₂} →          -- if M evaluates to (`suc V), then
                                           -- we evaluate the second branch M₂ and
    M ⇓ (`suc V) →                         -- return its value V₂, but since M₂ can
    M₂ [ x := ⌞ V ⌟ ] ⇓ V₂ →               -- use the variable x to refer to V,
    -------------------------------------  -- we first replace x with V in M₂.
    (case M [zero⇒ M₁ |suc x ⇒ M₂ ]) ⇓ V₂

  -- Finally the rule for μ-recursion is:

  ⇓-μ : ∀ {x M V} →              -- Recall, that in (μ x ⇒ M) the subterm M can
                                 -- use x to refer to (μ x ⇒ M) itself, just like
    M [ x := (μ x ⇒ M) ] ⇓ V →   -- in Agda we can define a function which
    --------------------------   -- recursively calls itself. Hence, we evaluate
    (μ x ⇒ M) ⇓ V                -- (μ x ⇒ M) by evaluating M in which we replaced
                                 -- x with (μ x ⇒ M) itself.

-- For the small-step semantics, we distinguished between β- and ξ-rules. The
-- β-rules describe the actual primitive computation steps, and the ξ-rules
-- describe how computation propagates through the subterms. Below you can see
-- the small-step rules for functions again:
--
--   β-ƛ :                               ξ-·₁ :                ξ-·₂ :
--                                                               Value M →
--     Value N →                           M —→ M′ →             N —→ N′ →
--     -----------------------------       ---------------       ---------------
--     (ƛ x ⇒ M) · N —→ M [ x := N ]       M · N —→ M′ · N       M · N —→ M · N′
--
-- If we have a program of the form (M · N) then a successful evaluation
-- proceeds as
--
--           (1)                (2)                (3)           (4)
--   (M · N) —↠ ((ƛ x ⇒ M) · N) —↠ ((ƛ x ⇒ M) · V) —→ M [x := V] —↠ V'
--
-- where the individual steps have the following form:
--
-- (1) 0 or more steps of the form (ξ-·₁ M—→M'), until M' is a value,
-- (2) 0 or more steps of the form (ξ-·₂ N—→N'), until N' is a value,
-- (3) Now that we have discovered the values of the subterms of (M · N),
--     we can use β-ƛ to perform the actual computation: the introduction form
--     for functions (ƛ x ⇒ M) is canceled out by the elimination form
--     for functions (application), which amounts to substituting
--     the argument value V for the paramter variable x in the function body M.
-- (4) Afterwards evaluation continues with the function body, until eventually
--     a final value V' is reached.
--
-- The rules of the big-step semantics, can be thought of as combinations
-- between β- and ξ-rules. Below you can see the big-step rule for functions:
--
--   ⇓-· : ∀ {x M M' N V V'} →
--
--     M ⇓ (ƛ x ⇒ M') →            -- this corresponds to (1),
--     N ⇓ V' →                    -- this corresponds to (2),
--     M' [ x := ⌞ V' ⌟ ] ⇓ V →    -- the substitution corresponds to (3) and
--     ------------------------    -- and the subsequent evaluation to V
--                                 -- corresponds to (4)
--     (M · N) ⇓ V
--
-- In the exercise about the equivalence of big-step and small-step semantics,
-- you will have to further investigate this correspondence.

variable
  V V' V₁ V₂ : Value
