module Exam.Exercise2.SoundnessSmallStep where

open import Exam.Exercise2.Syntax
open import Exam.Exercise2.Substitution
open import Exam.Exercise2.SemanticsSmallStep
open import Exam.Exercise2.Typing
open import Exam.Exercise2.Sublist
open import Exam.Exercise2.All2
open import Exam.Exercise2.SoundnessSubstitution

open import Relation.Binary.PropositionalEquality using (_≡_; _≢_; refl; sym)
open import Relation.Nullary                      using (Dec; yes; no; ¬_)
open import Data.String                           using (_≟_)
open import Data.List                             using (List; _∷_; []; length)
open import Data.Nat as Nat                       using (ℕ; zero; suc)
open import Data.Maybe                            using (Maybe; just; nothing)
open import Data.Product                          using (_×_; proj₁; proj₂; ∃-syntax) renaming (_,_ to ⟨_,_⟩)

-- Progress --------------------------------------------------------------------

-- Technically, we also need to prove the progress lemma to show type soundness
-- for our extended language. However, to keep the exam reasonable in size, we
-- decided to omit the progress lemma.

-- Preservation ----------------------------------------------------------------

-- Let's start by taking a look at the preservation theorem of our original
-- language:
--
--   preserve : ∀ {M N : Term} {T : Type} →
--     ∅ ⊢ M ⦂ T →
--     M —→ N →
--     ∅ ⊢ N ⦂ T
--
-- We somehow need to change this theorem, because both the typing and semantics
-- relations look different in our extended language: they talk about stores.
--
-- A first attempt might be to use the same approach as for contexts Γ, and
-- demand that the store and store typings are always empty:
--
--   preserve : ∀ {M N : Term} {T : Type} →
--     [] ; ∅ ⊢ M ⦂ T →
--     ⟨ [] , M ⟩→⟨ [] , N ⟩ →
--     ∅ ⊢ N ⦂ T
--
-- This approach worked for contexts, because if we start with a closed term,
-- then we never reduce it to a program which suddenly talks about a free
-- variable. We're only substituting closed terms into closed terms which again
-- yields closed terms.
--
-- However, this approach doesn't work for stores: while it is fine to assume
-- that we start with an empty store, the store is not necessarily going to stay
-- empty after reduction. This can be seen in the β-new rule, which extends the
-- store and returns a new location.
-- This new location corresponds to a free variable, in the sense that it refers
-- to external information (the store) which is not part of the program.
--
-- What we want is a preservation theorem which applys to all steps of an
-- evaluation, and the steps in the middle might already have some values in the
-- store.
--
-- So our next attempt looks as followed:
--
--   preserve : ∀ {Σ : StoreTy} {σ σ' : Store} {M N : Term} {T : Type} →
--     Σ ; ∅ ⊢ M ⦂ T →
--     ⟨ σ , M ⟩→⟨ σ' , N ⟩  →
--     ∃[ Σ' ] (Σ' ; ∅ ⊢ N ⦂ T)
--
-- Since we use an arbitrary store σ and store typing Σ, the lemma now also
-- works for the intermediary steps. That's good.
-- However, there is still something important missing:
-- The Σ from the typing (Σ ; ∅ ⊢ M ⦂ T) should in some sense match the σ from
-- the semantics ⟨ σ , M ⟩→⟨ σ' , N ⟩.
-- Similarly the resulting σ' should in some sense match the resulting store
-- typing Σ'.
--
-- This is where the All₂ type comes in. What we want to express is
--
--   if we have a value V at location ℓ in store σ,
--   then there should be a type A at location ℓ in Σ,
--   such that V has type A.
--
-- and we define it as followed:

infix 4 _⊢_

_⊢_ : StoreTy → Store → Set
Σ ⊢ σ = All₂ (λ T M → Σ ; ∅ ⊢ M ⦂ T) Σ σ
--                    ^––––––––––––––––––––––––––––––––––––––––––––––
--                                                                  |
-- An important detail is that the typing relation itself refers to Σ.
-- This is necessary as references may again contain other references, and there
-- can even be cycles of references.
-- So if we check that a value in our store σ has the corresponding type from
-- our store typing Σ, then this value may be another location, whose type we
-- need to lookup from the store typing again and so on.
--
-- Now were finally able to state the real preservation theorem:
--
--   preserve : ∀ {Σ : StoreTy} {σ σ' : Store} {M N : Term} {T : Type} →
--     Σ ⊢ σ →
--     Σ ; ∅ ⊢ M ⦂ T →
--     ⟨ σ , M ⟩→⟨ σ' , N ⟩  →
--     ∃[ Σ' ] (Σ' ⊢ σ' × Σ ⊆ Σ' × Σ' ; ∅ ⊢ N ⦂ T)
--
-- The theorem can then be read as followed:
--
--   If we have a store σ,
--   and a store typing Σ,
--   such that they fit together (Σ ⊢ σ),
--   and we have a program M which
--     - has type T for Σ, and
--     - together with σ evaluates to ⟨ σ' , N ⟩;
--   then there exists a store typing Σ',
--     - which still has the same types for the old locations (Σ ⊆ Σ'),
--     - but also fiting types for the new store σ'  (Σ' ⊢ σ'),
--     - and gives the resulting term N, the same type T as the original term M.


-- Lemmas you might find useful:

-- If M has type T using locations from the store typing Σ,
-- then M also has type T if we add additional locations to the store typing.
-- This is going to be important for the β-new rule.
postulate
  ⊆-preserves-⊢ : ∀ {Σ Σ' Γ} {M} {T} →
    Σ ⊆ Σ' →
    Σ  ; Γ ⊢ M ⦂ T →
    Σ' ; Γ ⊢ M ⦂ T
-- (The proof is straightforward, and in only one case you need a lemma, which
-- can be found in Sublist.agda.)

-- The next lemma unfortunately looks very intimidating. However, it has a very
-- short proof using only the previous lemma, and it arises naturally from the
-- subsequent lemma ⊢-ext.
-- This lemma basically says the same as the previous lemma, but instead for a
-- single typing (Σ ; Γ ⊢ M ⦂ T), it talks about many typings at once.
postulate
  ⊆-preserves-⊢* : ∀ {Σ Σ' Σ''} {σ} {M : Term} {T : Type} →
    Σ ⊆ Σ' →
    All₂ (λ T M → Σ  ; ∅ ⊢ M ⦂ T) Σ'' σ →
    All₂ (λ T M → Σ' ; ∅ ⊢ M ⦂ T) Σ'' σ

-- If we have a store σ with fitting store typing Σ,
-- then we can extend σ with a new term M, and Σ with a new type T,
-- if M has type T under Σ.
postulate
  ⊢-ext : ∀ {Σ} {σ} {T M} →
    Σ ; ∅ ⊢ M ⦂ T →
    Σ ⊢ σ →
    T ∷ Σ ⊢ M ∷ σ
-- (The proof only requires the ⊆-preserves-⊢ and ⊆-preserves-⊢* lemmas.)


-- The theorem you need to prove:

postulate
  preserve : ∀ {Σ : StoreTy} {σ σ' : Store} {M N : Term} {T : Type} →
    Σ ⊢ σ →
    Σ ; ∅ ⊢ M ⦂ T →
    ⟨ σ , M ⟩→⟨ σ' , N ⟩  →
    ∃[ Σ' ] (Σ' ⊢ σ' × Σ ⊆ Σ' × Σ' ; ∅ ⊢ N ⦂ T)

-- The proof for this theorem is by induction on ⟨ σ , M ⟩→⟨ σ' , N ⟩, and
-- requires a lot of lemmas:
--
-- - subst-preserve
-- - ⊆-refl
-- - ⊆-preserves-⊢
-- - all₂-length
-- - ⊢-ext
-- - lookup-length
-- - lookup-all
-- - update-all
--
-- However, it is not necessarily hard to prove, if you are able to identify the
-- right places to apply the lemmas.
